# -*- coding: utf-8 -*-
# modjkapi - API for modjk management
# Copyright © 2009-2011, Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
modjkapi main classes.
"""

__docformat__ = 'restructuredtext'
__version__ = '0.1.2.28'
__apiversion__ = '0.1.2.28'
__license__ = 'GPLv3+'
__author__ = "Steve 'Ashcrow' Milner"


import types
import warnings
import xml.sax

import urllib
import urllib2

from datetime import datetime

from modjkapi.mapper import JKManagerWorker, JKManagerBalancer, install_basic_auth_opener
from modjkapi.errors import JKCriteriaDoesNotExistError
from modjkapi.container import ObjectContainer


class JKManagerHandler(xml.sax.ContentHandler):
    """
    SAX XML handler for jkmanage.
    """

    meta_keywords = ['jk:status', 'jk:server', 'jk:software', 'jk:result']
    mod_jk_supported = ['mod_jk/1.2.28']

    def __init__(self, **kwargs):
        """
        Constructor for JKManagerHandler.

        :Parameters:
           - `kwargs`: all keyword arguments in dictionary format.
        """
        xml.sax.ContentHandler(**kwargs)
        self._current_balancer_name = None
        self._meta_info = {}
        self._balancers = {}

    def startElement(self, name, attrs):
        """
        Starts parsing at the start of an element.

        :Parameters:
           - `name`: the name of the element.
           - `attrs`: an attributes object.
        """
        # If the name is in the meta keywords then append to the
        # meta dictionary
        if name in self.meta_keywords:
            for attr_name in attrs.getNames():
                self._meta_info[attr_name.replace('jk:', '')] = attrs.get(
                    attr_name)
            # If we found a balancer then process it ...
        elif name == 'jk:balancer':
            data = {}
            balancer_name = attrs.get('name')
            self._current_balancer_name = balancer_name
            for item in attrs.items():
                data[item[0]] = item[1]
            self._balancers[balancer_name] = data
            if 'members' not in self._balancers[balancer_name].keys():
                self._balancers[balancer_name]['members'] = {}
        # If we find a member of a balancer then shove it in the balancer
        elif name == 'jk:member':
            data = {}
            balancer_name = self._current_balancer_name
            for item in attrs.items():
                data[item[0]] = item[1]
            self._balancers[balancer_name]['members'][attrs.get(
                'name')] = data

    def endElement(self, name):
        """
        Ends parsing of an element.

        :Parameters:
           - `name`: the name of the element.
        """
        # Warn the user if the version of mod_jk has not been verified
        if name == 'jk:software':
            if self._meta_info['jk_version'] not in self.mod_jk_supported:
                warnings.warn(
                    '%s is not in the supported mod_jk list! (%s)' % (
                        self._meta_info['jk_version'], self.mod_jk_supported))

    # Set up external properties read-only
    balancers = property(lambda self: self._balancers)
    meta_info = property(lambda self: self._meta_info)


class JKObjectFactory(ObjectContainer):
    """
    Parent class for JK object factories.
    """

    def __init__(self, jkmanager_xml_url, username=None, password=None, handler_class=JKManagerHandler):
        """
        Creates an JKObjectFactory instance.
           - `handler_class`: the handler class to use when creating objects.
           - `jkmanager_xml_url`: the url to snag the xml from.
        """
        self._handler_class = handler_class
        self._set_handler(self._handler_class())
        self.jkmanager_xml_url = jkmanager_xml_url
        self.username = username
        self.password = password
        self.parser = None
        self.last_time_updated = None
        if username and password:
            install_basic_auth_opener(self.jkmanager_xml_url, self.username, self.password)
        self.refresh()

    def _set_handler(self, handler):
        """
        Allows for setting the handler without forcing the factory
        out of sync.

        :Parameters:
           - `handler`: the new handler.
        """
        self._parsed = False
        self._handler = handler
        self._objects = []

    def _parse_structure(self):
        """
        Used to parse the structure of data.
        """
        raise exceptions.NotImplementedError(
            '_parse_structure must be defined.')


    def refresh(self):
        """
        Refreshes the data in the object pool.
        """
        self._handler = None
        self.parser = xml.sax.make_parser()
        self._set_handler(self._handler_class())
        self.parser.setContentHandler(self._handler)
        self.parser.parse(self.jkmanager_xml_url)
        self._parsed = False
        self.last_time_updated = datetime.now()

    def objects(self):
        """
        Override ObjectContainer.objects(). Generates and returns the objects.
        """
        self._parse_structure()
        return self._objects

    # Set up external properties
    handler = property(lambda self: self._handler,
                       lambda self: self._set_handler)


class JKManagerWorkerObjectFactory(JKObjectFactory):
    """
    Creates worker objects out of a parsed handler object.
    """

    def __init__(self, jkmanager_xml_url, handler_class=JKManagerHandler):
        """
        Creates the factory with a reference to the handler.

        :Parameters:
           - `handler_class`: the handler class to use when creating objects.
           - `jkmanager_xml_url`: the url to snag the xml from.
        """
        JKObjectFactory.__init__(self, jkmanager_xml_url, handler_class)
        self._meta_info = self._handler.meta_info
        self._balancers = self._handler.balancers

    def _parse_structure(self):
        """
        Parse the structure from a JKManagerHandler.
        """
        if not self._parsed:
            for balancer_name in self._balancers.keys():
                for worker in self._balancers[balancer_name]['members'].keys():
                    data = self._balancers[balancer_name]['members'][worker]
                    self._objects.append(
                        JKManagerWorker(worker,
                                        self._meta_info,
                                        data,
                                        self._balancers[balancer_name]))
            self._parsed = True


class JKManagerBalancerObjectFactory(JKObjectFactory):
    """
    Creates balancer objects out of a parsed handler object.
    """

    def __init__(self, jkmanager_xml_url, username=None, password=None, handler_class=JKManagerHandler):
        """
        Creates the factory with a reference to the handler.

        :Parameters:
           - `handler_class`: the handler class to use when creating objects.
           - `jkmanager_xml_url`: the url to snag the xml from.
        """
        JKObjectFactory.__init__(self, jkmanager_xml_url, username, password, handler_class)
        self._meta_info = self._handler.meta_info
        self._balancers = self._handler.balancers

    def _parse_structure(self):
        """
        Parse the structure from a JKManagerHandler.
        """
        if not self._parsed:
            for balancer_name in self._balancers.keys():
                worker_list = []
                data = self._balancers[balancer_name]
                for worker in self._balancers[balancer_name]['members'].keys():
                    worker_data = \
                        self._balancers[balancer_name]['members'][worker]
                    worker_list.append(JKManagerWorker(worker,
                                       self._meta_info,
                                       worker_data,
                                       self._balancers[balancer_name], 
                                       self.username, 
                                       self.password))
                self._objects.append(JKManagerBalancer(balancer_name,
                                                       self._meta_info,
                                                       data,
                                                       worker_list))
            self._parsed = True
