# -*- coding: utf-8 -*-
# modjkapi - API for modjk management
# Copyright © 2009-2011, Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Main classes for modjkapi.
"""

__docformat__ = 'restructuredtext'

from types import *

import urllib
import urllib2

import xml.sax

from modjkapi.errors import (JKResultError,
                             WorkerNotFoundError,
                             JKCriteriaDoesNotExistError)
from modjkapi.container import ObjectContainer

class HTTPBasicAuthPasswordManagerOpener(urllib.FancyURLopener):
    """
    A simple URL opener that holds a username & password for
    basic authentication requests.
    """

    def add_password(self, realm, host, username, password):
        """
        Add a password to the password cache
        """
        #print "HTTPBasicAuthPasswordManagerOpener::add_password(realm=%s, host=%s, username=%s, password=%s)" % (realm, host, username, password)
        key = host.lower()
        if realm:
            key = realm + '@' + key
        self.auth_cache[key] = (username, password)
        #print "HTTPBasicAuthPasswordManagerOpener::add_password: auth_cache=%s" % self.auth_cache

    def get_user_passwd(self, host, realm, clear_cache = 0):
        """
        Get the a username and password for a realm+host combination.
        Allow the realm to be None.
        """
        #print "HTTPBasicAuthPasswordManagerOpener::GET_USER_PASSWORD(host=%s, realm=%s)" % (host, realm)
        #print "HTTPBasicAuthPasswordManagerOpener::GET_USER_PASSWORD: auth_cache=%s" % self.auth_cache
        user, passwd = urllib.FancyURLopener.get_user_passwd(self, realm, host, clear_cache)
        if not (user or passwd):
            #print "HTTPBasicAuthPasswordManagerOpener::GET_USER_PASSWORD...trying realm=None"
            key = host.lower()
            if key in self.auth_cache:
                if clear_cache:
                    del self.auth_cache[key]
                else:
                    user, passwd = self.auth_cache[key]
        #print "HTTPBasicAuthPasswordManagerOpener::GET_USER_PASSWORD: user=%s, passwd=%s" % (user, passwd)
        return user, passwd

    def prompt_user_passwd(self, host, realm):
        """
        Override to return None, None always.  Since, in this context
        we never want to prompt for the password.
        """
        return None, None

def install_basic_auth_opener(jkmanager_xml_url, username, password):
    """
    Install the opener to handle basic auth
    """
    #print "install_basic_auth_opener(jkmanager_xml_url=%s, username=%s, password=%s)" % (jkmanager_xml_url, username, password)

    # Install the opener for urllib2
    password_mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
    password_mgr.add_password(None, jkmanager_xml_url, username, password)
    handler = urllib2.HTTPBasicAuthHandler(password_mgr)
    urllib2_opener = urllib2.build_opener(handler)
    urllib2.install_opener(urllib2_opener)

    # Install the opener for urllib 
    # in case urllib is used
    urllib_opener = urllib._urlopener
    if not (urllib_opener or isinstance(urllib_opener, HTTPBasicAuthPasswordManagerOpener)):
        urllib_opener = HTTPBasicAuthPasswordManagerOpener()
    type, url = urllib.splittype(jkmanager_xml_url)
    host, uri = urllib.splithost(url)
    host, port = urllib.splitport(host)
    urllib_opener.add_password(None, host, username, password)
    urllib._urlopener = urllib_opener

class JKMapper(ObjectContainer):
    """
    Master class for mappers.
    """

    def _get_protocol(self, ssl):
        """
        Returns the correct protocol based on ssl.

        :Parameters:
           - `ssl`: sets if ssl should be used.
        """
        ObjectContainer.__init__(self)
        protocol = 'https'
        if not ssl:
            protocol = 'http'
        return protocol

    def _get_and_parse_response(self, url, username=None, password=None):
        """
        Returns the correct protocol based on ssl.

        :Parameters:
           - `url`: url to connect.
        """

        class JKResultHandler(xml.sax.ContentHandler):
            """
            Simple result handler for modjk.
            """

            def startElement(self, name, attrs):
                """
                Starts parsing at the start of an element.

                :Parameters:
                   - `name`: the name of the element.
                   - `attrs`: an attributes object.
                """
                if name == 'jk:result':
                    if attrs.get('type').lower() != 'ok':
                        raise JKResultError(attrs.get('message'))

        # Connect and get shove the data into the parser.
        
        conn = urllib.urlopen(url)
        handler = JKResultHandler()
        parser = xml.sax.make_parser()
        parser.setContentHandler(handler)
        parser.parse(conn)
        conn.close()


class JKManagerBalancer(JKMapper):
    """
    A representation of the JKManager Balancer.
    """

    def __init__(self, name, meta_info, data, workers=[]):
        """
        Creates the balancer representation.

        :Parameters:
           - `name`: the name of the worker.
           - `meta_info`: the dictionary meta information for the mapper.
           - `data`: dictionary that represent the balancer directly.
           - `workers`: list of workers that belond to the balancer.
        """
        JKMapper.__init__(self)
        self.name = name
        self.meta = meta_info
        for key in data.keys():
            self.__setattr__(key, data[key])
        self._objects = workers

        # Map some methods to keep backwards compat
        self.filter_workers_by = self.filter_by
        self.get_one_worker = self.get_one

    def __repr__(self):
        """
        String representation of the object.
        """
        return '<JKManagerBalancer "%s">' % self.name

    def get_worker(self, name):
        """
        :Parameters:
           - `name`: the name of the worker being looked for.
        """
        if name in self._objects.keys():
            return self._objects[key]
        raise WorkerNotFoundError('%s is not defined.' % key)

    # Read only properties
    workers = property(lambda self: self._objects)


class JKManagerWorker(JKMapper):
    """
    A representation of the JKManager Worker.
    """

    WORKER_ACTIVATION_STOPPED = "2"
    WORKER_ACTIVATION_ENABLED = "0"
    WORKER_ACTIVATION_DISABLED = "1"

    def __init__(self, name, meta_info, data, balancer, username=None, password=None):
        """
        Creates the worker representation.

        :Parameters:
           - `name`: the name of the worker.
           - `meta_info`: the dictionary meta information for the worker.
           - `data`: dictionary that represent the worker directly.
           - `balancer`: data from the balancer. Simple housing.
        """
        JKMapper.__init__(self)
        self.name = name
        self.meta = meta_info
        self._objects = []
        self.balancer = balancer
        for key in data.keys():
            self.__setattr__(key, data[key])

        self._url = "%s/jkmanage?cmd=update&w=%s&sw=%s&mime=xml" % (
                        self.meta['name'],
                        self.balancer['name'],
                        self.name)

        self.username = username
        self.password = password

        if username and password:
            full_url = "%s://%s" % (self._get_protocol(False), self._url)
            install_basic_auth_opener(full_url, self.username, self.password)


    def __repr__(self):
        """
        String representation of the object.
        """
        return '<JKManagerWorker "%s" for "%s">' % (self.name,
                                                    self.meta['name'])

    def _edit_worker(self, worker_activation, ssl=True):
        """
        Edit a workers properties

        :Parameters:
            -`ssl`: sets if ssl should be used.
        """
        protocol = self._get_protocol(ssl)
        url = "%s://%s&%s" % (protocol, self._url, "vwa=%s&wa=%s" % (worker_activation,worker_activation))
        self._get_and_parse_response(url, self.username, self.password)


    def stop(self, ssl=True):
        """
        Stop a worker

        :Parameters:
            - `ssl`: sets if ssl should be used.
        """
        self._edit_worker(JKManagerWorker.WORKER_ACTIVATION_STOPPED, ssl) 

    def disable(self, ssl=True):
        """
        Disable a worker

        :Parameters:
           - `ssl`: sets if ssl should be used.
        """
        self._edit_worker(JKManagerWorker.WORKER_ACTIVATION_DISABLED, ssl)


    def enable(self, ssl=True):
        """
        Enable a worker

        :Parameters:
           - `ssl`: sets if ssl should be used.
        """
        self._edit_worker(JKManagerWorker.WORKER_ACTIVATION_ENABLED, ssl)
